package edu.kit.ipd.milton;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.io.FileUtils;

import edu.kit.ipd.milton.events.AssociatedEventCollection;
import edu.kit.ipd.milton.events.Change;
import edu.kit.ipd.milton.events.Event;
import edu.kit.ipd.milton.events.EventType;
import edu.kit.ipd.milton.events.Normalization;
import edu.kit.ipd.milton.events.Pattern;
import edu.kit.ipd.milton.measure.Measure;
import edu.kit.ipd.milton.measure.MeasureValue;
import edu.kit.ipd.milton.pairing.EventPairer;
import edu.kit.ipd.milton.pairing.OsbAdapter;
import edu.kit.ipd.milton.pairing.PairedEventDistanceFunction;
import edu.kit.ipd.milton.timeseries.TimeSeries;
import edu.kit.ipd.milton.timeseries.TimeSeriesPoint;
import edu.kit.ipd.milton.util.IntervalNormalizationOption;
import edu.kit.ipd.milton.util.Tuple;

/**
 * 
 * @author aenglhardt
 *
 */
public class Milton {
	public static final double DEFAULT_WARPWIN = 0.1;

	/**
	 * 
	 * @param path
	 *            path to the time series files
	 * @param fileA
	 *            original time series file
	 * @param fileB
	 *            changed time series file
	 * @param measure
	 *            parameterized measure
	 * @param normOptions
	 *            normalization options
	 * @param eventType
	 *            what event type is used
	 * @param warpWin
	 *            length of maximum window for pairing (relative (0..1])
	 * @param timeInterval
	 *            start and end timestamp of the time series
	 * @param chattyOutput
	 *            whether to verbose output during calculation
	 * @return calculated error
	 */
	public static MeasureValue runMeasure(String path, String fileA, String fileB, Measure measure,
			EventType eventType, IntervalNormalizationOption normOptions, double warpWin, Tuple<Double> timeInterval,
			boolean chattyOutput) {
		final TimeSeries originalTimeSeries = new TimeSeries(path + fileA);
		final TimeSeries changedTimeSeries = new TimeSeries(path + fileB);

		chattyOutput(originalTimeSeries, chattyOutput);
		chattyOutput(changedTimeSeries, chattyOutput);

		List<Event> originalEvents;
		List<Event> changedEvents;
		if (eventType == EventType.CHANGE) {
			originalEvents = Change.calculateFromTimeSeries(originalTimeSeries);
			changedEvents = Change.calculateFromTimeSeries(changedTimeSeries);
		} else if (eventType == EventType.PATTERN) {
			originalEvents = Pattern.calculateFromTimeSeries(originalTimeSeries);
			changedEvents = Pattern.calculateFromTimeSeries(changedTimeSeries);
		} else {
			throw new RuntimeException("Unsupported EventType");
		}

		chattyOutput("Events:", chattyOutput);
		chattyOutput(originalEvents, chattyOutput);
		chattyOutput(changedEvents, chattyOutput);

		final PairedEventDistanceFunction distFn = new PairedEventDistanceFunction(measure.getErrorTIME(),
				measure.getErrorSCORE());
		final EventPairer pairer = new OsbAdapter(distFn, warpWin);

		handleNormalization(originalEvents, changedEvents, eventType, normOptions, pairer, true);
		rebuildTimeSeries(originalEvents, originalTimeSeries.getPoints(), eventType);
		rebuildTimeSeries(changedEvents, changedTimeSeries.getPoints(), eventType);

		MeasureValue value;
		if (originalTimeSeries.size() == 0 && changedTimeSeries.size() == 0) {
			System.out.println("No events found");
			value = MeasureValue.defaultError();
		} else {
			handleNormalization(originalEvents, changedEvents, eventType, normOptions, pairer, false);
			AssociatedEventCollection splitChanges;
			if (originalEvents.size() == 0 || changedEvents.size() == 0) {
				splitChanges = new AssociatedEventCollection();
				splitChanges.buildFromMissingChanges(originalEvents, changedEvents);
			} else {
				splitChanges = pairer.matchEvents(originalEvents, changedEvents, timeInterval);
			}
			chattyOutput(splitChanges, chattyOutput);
			value = measure.calcMeasure(splitChanges);
		}
		chattyOutput(value, chattyOutput);
		return value;
	}

	/**
	 * Exports a measure to a given path
	 * 
	 * @param pathName
	 * @param fileA
	 * @param fileB
	 * @param value
	 *            measure error
	 */
	public static void exportMeasure(String pathName, String fileA, String fileB, MeasureValue value) {
		try {
			final String fileName = pathName + fileA + "_" + fileB + ".err";
			FileUtils.writeStringToFile(new File(fileName), value.toString());
		} catch (IOException e) {
			throw new RuntimeException("Failed exporting measure!");
		}
	}

	public static final void handleNormalization(List<Event> originalEvents, List<Event> changedEvents,
			EventType eventType, IntervalNormalizationOption normOptions, EventPairer pairer, boolean pre) {
		if (pairer == null || pre && pairer.preNormalize() || !pre && !pairer.preNormalize()) {
			if (eventType == EventType.CHANGE) {
				List<Change> castedListA = new LinkedList<>();
				List<Change> castedListB = new LinkedList<>();
				for (Event e : originalEvents) {
					castedListA.add((Change) e);
				}
				for (Event e : changedEvents) {
					castedListB.add((Change) e);
				}
				Tuple<List<Change>> resultTuple = normalizeChanges(castedListA, castedListB, normOptions);
				castedListA = resultTuple.getA();
				castedListB = resultTuple.getB();
				originalEvents.clear();
				changedEvents.clear();
				for (Change c : castedListA) {
					originalEvents.add(c);
				}
				for (Change c : castedListB) {
					changedEvents.add(c);
				}
			} else if (eventType == EventType.PATTERN) {
				List<Pattern> castedListA = new LinkedList<>();
				List<Pattern> castedListB = new LinkedList<>();
				for (Event e : originalEvents) {
					castedListA.add((Pattern) e);
				}
				for (Event e : changedEvents) {
					castedListB.add((Pattern) e);
				}
				Tuple<List<Pattern>> resultTuple = normalizePatterns(castedListA, castedListB, normOptions);
				castedListA = resultTuple.getA();
				castedListB = resultTuple.getB();
				originalEvents.clear();
				changedEvents.clear();
				for (Pattern c : castedListA) {
					originalEvents.add(c);
				}
				for (Pattern c : castedListB) {
					changedEvents.add(c);
				}
			} else {
				throw new RuntimeException("Unsupported EventType");
			}
		}

	}

	public static final Tuple<List<Change>> normalizeChanges(List<Change> castedListA, List<Change> castedListB,
			IntervalNormalizationOption normOptions) {
		if (normOptions.timeIntervalNormalization()) {
			castedListA = Normalization.normalizeTimeTimeSeriesWithChangeInterval(castedListA, normOptions
					.getTimeInterval().getA().doubleValue(), normOptions.getTimeInterval().getB().doubleValue(),
					normOptions.getEpsilon());
			castedListB = Normalization.normalizeTimeTimeSeriesWithChangeInterval(castedListB, normOptions
					.getTimeInterval().getA().doubleValue(), normOptions.getTimeInterval().getB().doubleValue(),
					normOptions.getEpsilon());
		} else {
			Tuple<List<Change>> resultTuple = Normalization.normalizeTimeTimeSeriesOnlyOnChanges(castedListA,
					castedListB, normOptions.getEpsilon());
			castedListA = resultTuple.getA();
			castedListB = resultTuple.getB();

		}
		if (normOptions.scoreIntervalNormalization()) {
			castedListA = Normalization.normalizeScoreTimeSeriesWithChangeInterval(castedListA, normOptions
					.getScoreInterval().getA().doubleValue(), normOptions.getScoreInterval().getB().doubleValue(),
					normOptions.getEpsilon());
			castedListB = Normalization.normalizeScoreTimeSeriesWithChangeInterval(castedListB, normOptions
					.getScoreInterval().getA().doubleValue(), normOptions.getScoreInterval().getB().doubleValue(),
					normOptions.getEpsilon());
		} else {
			Tuple<List<Change>> resultTuple = Normalization.normalizeScoreTimeSeriesOnlyOnChanges(castedListA,
					castedListB, normOptions.getEpsilon());
			castedListA = resultTuple.getA();
			castedListB = resultTuple.getB();
		}

		return new Tuple<List<Change>>(castedListA, castedListB);
	}

	public static final Tuple<List<Pattern>> normalizePatterns(List<Pattern> castedListA, List<Pattern> castedListB,
			IntervalNormalizationOption normOptions) {
		if (normOptions.timeIntervalNormalization()) {
			if (castedListA != null)
				castedListA = Normalization.normalizeLocationTimeSeriesPatternsWithInterval(castedListA, normOptions
						.getTimeInterval().getA().doubleValue(), normOptions.getTimeInterval().getB().doubleValue(),
						normOptions.getEpsilon());
			if (castedListB != null)
				castedListB = Normalization.normalizeLocationTimeSeriesPatternsWithInterval(castedListB, normOptions
						.getTimeInterval().getA().doubleValue(), normOptions.getTimeInterval().getB().doubleValue(),
						normOptions.getEpsilon());
		} else {
			Tuple<List<Pattern>> resultTuple = Normalization.normalizeLocationPatternsOnly(castedListA, castedListB,
					normOptions.getEpsilon());
			castedListA = resultTuple.getA();
			castedListB = resultTuple.getB();

		}
		if (normOptions.scoreIntervalNormalization()) {
			if (castedListA != null)
				castedListA = Normalization.normalizeScoreTimeSeriesPatternsWithInterval(castedListA, normOptions
						.getScoreInterval().getA().doubleValue(), normOptions.getScoreInterval().getB().doubleValue(),
						normOptions.getEpsilon());
			if (castedListB != null)
				castedListB = Normalization.normalizeScoreTimeSeriesPatternsWithInterval(castedListB, normOptions
						.getScoreInterval().getA().doubleValue(), normOptions.getScoreInterval().getB().doubleValue(),
						normOptions.getEpsilon());
		} else {
			Tuple<List<Pattern>> resultTuple = Normalization.normalizeScorePatternsOnly(castedListA, castedListB,
					normOptions.getEpsilon());
			castedListA = resultTuple.getA();
			castedListB = resultTuple.getB();
		}

		return new Tuple<List<Pattern>>(castedListA, castedListB);
	}

	public static final void rebuildTimeSeries(List<Event> events, List<TimeSeriesPoint> tsArray, EventType eventType) {
		tsArray.clear();
		if (eventType == EventType.CHANGE) {
			for (Event e : events) {
				tsArray.add(new TimeSeriesPoint(((Change) e).getTimestamp(), new double[] { ((Change) e).getScore() }));
			}
		} else if (eventType == EventType.PATTERN) {
			for (Event e : events) {
				tsArray.add(new TimeSeriesPoint(((Pattern) e).getLocation(), ((Pattern) e).getEventSeries()));
			}
		} else {
			throw new RuntimeException("Unsupported EventType");
		}

	}

	private static void chattyOutput(Object o, boolean chatty) {
		if (chatty)
			System.out.println(o);
	}

	private Milton() {

	}

}
