package edu.kit.ipd.milton.algorithms.changedetection;

import java.util.LinkedList;
import java.util.List;

import edu.kit.ipd.milton.events.Change;

/**
 * 
 * @author aenglhardt
 *
 */
public class Cusum implements ChangeDetection {
	private double epsModifier;
	private double epsilon;
	private double drift;

	/**
	 * 
	 * @param epsilon
	 *            cusum parameter
	 * @param drift
	 *            drift to ignore
	 */
	public Cusum(double epsilon, double drift) {
		this.epsModifier = Double.MIN_VALUE;
		this.epsilon = epsilon;
		this.drift = drift;
	}

	/**
	 * 
	 * @param epsilonModifier
	 *            modifier which is multiplied to each (maximum - minimum) of a
	 *            time series for calculate an epsilon value
	 */
	public Cusum(double epsilonModifier) {
		this.epsModifier = epsilonModifier;
		this.epsilon = 0;
		this.drift = 0;
	}

	@Override
	public void setNewParam(double[][] ts) {
		if (epsModifier == Double.MIN_VALUE)
			return;

		double max = Double.MIN_VALUE;
		double min = Double.MAX_VALUE;
		for (int i = 0; i < ts.length; i++) {
			if (ts[i][1] > max)
				max = ts[i][1];
			if (ts[i][1] < min)
				min = ts[i][1];
		}
		this.epsilon = (max - min) * epsModifier;
	}

	/**
	 * 
	 * @param ts
	 *            time series array with time in [i][0] and value in [i][1]
	 * @return calculated changes
	 */
	@Override
	public List<Change> detectChanges(double[][] ts) {
		List<Change> changePoints = new LinkedList<>();
		final int length = ts.length;
		final double[] s_p = new double[length];
		final double[] s_m = new double[length];
		final double[] s_n = new double[length];

		boolean foundCP = false;
		double diff = 0;
		double score;
		for (int i = 1; i < length; i++) {
			diff = ts[i][1] - ts[i - 1][1];
			if (!foundCP) {
				s_p[i] = Math.max(0, s_p[i - 1] + diff - drift);
				s_m[i] = Math.max(0, s_m[i - 1] - diff - drift);
			} else {
				s_p[i] = Math.max(0, 0 + diff - drift);
				s_m[i] = Math.max(0, 0 - diff - drift);
				foundCP = false;
			}

			s_n[i] = s_n[i - 1] + diff - drift;
			if (s_p[i] > epsilon || s_m[i] > epsilon) {
				score = Math.max(s_p[i], s_m[i]);
				changePoints.add(new Change(ts[i][0], score));
				foundCP = true;
			}
		}
		return changePoints;
	}
}
