package edu.kit.ipd.milton.pairing;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import edu.kit.ipd.milton.events.AssociatedEventCollection;
import edu.kit.ipd.milton.events.Change;
import edu.kit.ipd.milton.events.Event;
import edu.kit.ipd.milton.events.EventType;
import edu.kit.ipd.milton.events.Normalization;
import edu.kit.ipd.milton.events.Pattern;
import edu.kit.ipd.milton.util.Tuple;
import edu.kit.ipd.osb.OSB;
import edu.kit.ipd.osb.OSB.OSBResult;

/**
 * 
 * @author aenglhardt
 *
 */
public class OsbAdapter implements EventPairer {

	private final PairedEventDistanceFunction distFn;
	private final double warpWin;

	/**
	 * 
	 * @param distFn
	 * @param warpWin
	 *            percentage that is multiplied with input time interval and
	 *            calculates the maximum pair distance.
	 */
	public OsbAdapter(PairedEventDistanceFunction distFn, double warpWin) {
		this.distFn = distFn;
		this.warpWin = warpWin;
	}

	@Override
	public AssociatedEventCollection matchEvents(List<Event> originalEvents, List<Event> changedEvents,
			Tuple<Double> timeInterval) {
		// Lists are cloned during normalization
		Tuple<List<Event>> normalizedTuple = normalizeBothForPairer(originalEvents, changedEvents);
		List<Event> originalEventsClone = normalizedTuple.getA();
		List<Event> changedEventsClone = normalizedTuple.getB();

		final int m = originalEventsClone.size();
		final int n = changedEventsClone.size();
		final double[][] distMtx = calcDistancematrix(originalEventsClone, changedEventsClone, m, n);
		final double jumpcost = OSB.estimateJumpcosts(distMtx);
		final int queryskip = Math.max(originalEvents.size(), changedEvents.size());
		final int targetskip = Math.max(originalEvents.size(), changedEvents.size());

		OSBResult matching = OSB.osb(distMtx, m, n, Integer.MAX_VALUE, queryskip, targetskip, jumpcost);

		final List<Tuple<Event>> pairedChanges = new LinkedList<>();
		final List<Event> missingChanges = new LinkedList<>();
		final List<Event> falsePositiveChanges = new LinkedList<>();

		final Set<Integer> pairedIs = new HashSet<>();
		final Set<Integer> pairedJs = new HashSet<>();
		final double curWarpWin = warpWin
				* Math.abs(timeInterval.getB().doubleValue() - timeInterval.getA().doubleValue());
		for (int i = 0; i < matching.indxRow.length; i++) {
			double curDist = originalEvents.get(matching.indxRow[i]).calcDistance(
					changedEvents.get(matching.indxCol[i]));
			if (warpWin == -1 || curDist <= curWarpWin) {
				pairedChanges.add(new Tuple<Event>(originalEvents.get(matching.indxRow[i]), changedEvents
						.get(matching.indxCol[i])));
				pairedIs.add(Integer.valueOf(matching.indxRow[i]));
				pairedJs.add(Integer.valueOf(matching.indxCol[i]));
			}
		}

		for (int i = 0; i < originalEvents.size(); i++) {
			if (!pairedIs.contains(Integer.valueOf(i)))
				missingChanges.add(originalEvents.get(i));
		}
		for (int j = 0; j < changedEvents.size(); j++) {
			if (!pairedJs.contains(Integer.valueOf(j)))
				falsePositiveChanges.add(changedEvents.get(j));
		}

		return new AssociatedEventCollection(pairedChanges, missingChanges, falsePositiveChanges);
	}

	private double[][] calcDistancematrix(List<Event> originalEvents, List<Event> changedEvents, int m, int n) {
		final double[][] distMtx = new double[m][n];
		for (int i = 0; i < m; i++)
			for (int j = 0; j < n; j++)
				distMtx[i][j] = distFn.calcDistance(originalEvents.get(i), changedEvents.get(j));

		return distMtx;
	}

	public static Tuple<List<Event>> normalizeBothForPairer(List<Event> a, List<Event> b) {
		if (a.size() < 1 && b.size() < 1)
			return new Tuple<List<Event>>(new LinkedList<Event>(), new LinkedList<Event>());
		if (a.get(0).getType() == EventType.CHANGE) {
			List<Change> castedListA = new LinkedList<>();
			List<Change> castedListB = new LinkedList<>();
			for (Event e : a) {
				castedListA.add((Change) e);
			}
			for (Event e : b) {
				castedListB.add((Change) e);
			}
			Tuple<List<Change>> result = Normalization.normalizeScoreTimeSeriesOnlyOnChanges(castedListA, castedListB,
					Normalization.EPSILON);

			result = Normalization.normalizeTimeTimeSeriesOnlyOnChanges(result.getA(), result.getB(),
					Normalization.EPSILON);
			List<Event> castedListEventA = new LinkedList<>();
			List<Event> castedListEventB = new LinkedList<>();
			for (Change c : result.getA()) {
				castedListEventA.add(c);
			}
			for (Change c : result.getB()) {
				castedListEventB.add(c);
			}
			return new Tuple<>(castedListEventA, castedListEventB);
		} else if (a.get(0).getType() == EventType.PATTERN) {
			List<Pattern> castedListA = new LinkedList<>();
			List<Pattern> castedListB = new LinkedList<>();
			for (Event e : a) {
				castedListA.add((Pattern) e);
			}
			for (Event e : b) {
				castedListB.add((Pattern) e);
			}
			Tuple<List<Pattern>> result = Normalization.normalizeLocationAndScorePatternsOnly(castedListA, castedListB,
					Normalization.EPSILON);
			List<Event> castedListEventA = new LinkedList<>();
			List<Event> castedListEventB = new LinkedList<>();
			for (Pattern c : result.getA()) {
				castedListEventA.add(c);
			}
			for (Pattern c : result.getB()) {
				castedListEventB.add(c);
			}
			return new Tuple<>(castedListEventA, castedListEventB);
		} else {
			throw new RuntimeException("Cannot normalize unknown type for pairing");
		}
	}

	@Override
	public boolean preNormalize() {
		return false;
	}
}
