package edu.kit.ipd.osb;

import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * 
 * @author aenglhardt
 *
 */
public class OSB {

	public static final double JUMPCOST = 1;
	private static final double MIN_JUMPCOSTS = 0.0001;

	private OSB() {

	}

	public static class OSBResult {
		public final double pathcost;
		public final int[] indxRow;
		public final int[] indxCol;

		public OSBResult(double pathcost, int[] indxRow, int[] indxCol) {
			this.pathcost = pathcost / indxRow.length;
			this.indxRow = indxRow;
			this.indxCol = indxCol;
		}

		@Override
		public String toString() {
			return "OSBResult [pathcost=" + pathcost + ", indxRow=" + Arrays.toString(indxRow) + ", indxCol="
					+ Arrays.toString(indxCol) + "]";
		}

	}

	/**
	 * 
	 * @param t1
	 *            first time series
	 * @param t2
	 *            second time series
	 * @param s1
	 *            score of first time series
	 * @param s2
	 *            score of second time series
	 * @param warpwin
	 *            the warping window, same as that used in Dynamic Time Warping
	 * @param queryskip
	 *            sets a restriction on how many consecutive elements of
	 *            sequence t1 can be skipped in the matching
	 * @param targetskip
	 *            sets a restriction on how many consecutive elements of
	 *            sequence t2 can be skipped in the matching
	 * @param jumpcost
	 *            the cost of jumps in t1 or t2 used to balance the similarity
	 *            of elements. The number of jumps is multiplied by jumpcost. if
	 *            jumpcost == -1, then it will be estimated
	 * @return
	 */
	public static OSBResult osb(double[] t1, double[] t2, double[] s1, double[] s2, int warpwin, int queryskip,
			int targetskip, double jumpcost) {
		int m = t1.length;
		int n = t2.length;
		double[][] matx = new double[m][n];
		for (int i = 0; i < m; i++) {
			for (int j = 0; j < n; j++) {
				matx[i][j] = Math.abs((t2[j] - t1[i])) + Math.abs((s2[j] - s1[i]));
			}
		}

		if (jumpcost == -1) {
			jumpcost = estimateJumpcosts(matx);
		}

		double[][] matxE = extendMatrix(matx, m, n);

		return findpathDAG(matxE, warpwin, queryskip, targetskip, jumpcost);
	}

	/**
	 * 
	 * @param matx
	 *            distance matrix
	 * @param warpwin
	 *            the warping window, same as that used in Dynamic Time Warping
	 * @param queryskip
	 *            sets a restriction on how many consecutive elements of
	 *            sequence t1 can be skipped in the matching
	 * @param targetskip
	 *            sets a restriction on how many consecutive elements of
	 *            sequence t2 can be skipped in the matching
	 * @param jumpcost
	 *            the cost of jumps in t1 or t2 used to balance the similarity
	 *            of elements. The number of jumps is multiplied by jumpcost. if
	 *            jumpcost == -1, then it will be estimated
	 * @return
	 */
	public static OSBResult osb(double[][] matx, int m, int n, int warpwin, int queryskip, int targetskip,
			double jumpcost) {
		if (jumpcost == -1) {
			jumpcost = estimateJumpcosts(matx);
		}

		double[][] matxE = extendMatrix(matx, m, n);

		return findpathDAG(matxE, warpwin, queryskip, targetskip, jumpcost);
	}

	private static OSBResult findpathDAG(double[][] matx, int warpwin, int queryskip, int targetskip, double jumpcost) {
		int m = matx.length;
		int n = matx[0].length;
		double[][] weight = new double[m][n];
		for (int i = 0; i < m; i++) {
			Arrays.fill(weight[i], Double.MAX_VALUE);
		}
		int[][] camefromrow = new int[m][n];
		int[][] camefromcol = new int[m][n];

		for (int j = 0; j < n; j++) {
			weight[0][j] = matx[0][j];
		}
		for (int i = 0; i < m; i++) {
			weight[i][0] = matx[i][0];
		}

		for (int i = 0; i < m - 1; i++) {
			for (int j = 0; j < n - 1; j++) {
				if (Math.abs(i - j) <= warpwin) {
					int stoprowjump = Math.min(m - 1, i + queryskip);
					for (int rowjump = i + 1; rowjump <= stoprowjump; rowjump++) {
						int stopk = Math.min(n - 1, j + targetskip);
						for (int k = j + 1; k <= stopk; k++) {
							double newweight = weight[i][j] + matx[rowjump][k] + ((rowjump - i - 1) + (k - j - 1))
									* jumpcost;
							if (weight[rowjump][k] > newweight) {
								weight[rowjump][k] = newweight;
								camefromrow[rowjump][k] = i;
								camefromcol[rowjump][k] = j;
							}
						}
					}
				}
			}
		}

		double pathcost = weight[m - 1][n - 1];
		int mincol = n - 1;
		int minrow = m - 1;
		List<Integer> indxcol = new LinkedList<>();
		List<Integer> indxrow = new LinkedList<>();
		while (mincol > 0 && minrow > 0) {
			indxcol.add(Integer.valueOf(mincol));
			indxrow.add(Integer.valueOf(minrow));
			int mincoltemp = camefromcol[minrow][mincol];
			minrow = camefromrow[minrow][mincol];
			mincol = mincoltemp;
		}

		int numIndx = indxcol.size();
		int[] indxcolArr = new int[numIndx - 1];
		int[] indxrowArr = new int[numIndx - 1];
		Collections.reverse(indxcol);
		Collections.reverse(indxrow);
		for (int i = 0; i < numIndx - 1; i++) {
			indxcolArr[i] = indxcol.get(i).intValue() - 1;
			indxrowArr[i] = indxrow.get(i).intValue() - 1;
		}

		return new OSBResult(pathcost, indxrowArr, indxcolArr);
	}

	public static double estimateJumpcosts(double[][] distMatrix) {
		if (distMatrix.length < 1 || distMatrix[0] == null)
			return JUMPCOST;

		double[] minDist = new double[distMatrix.length];
		for (int i = 0; i < distMatrix.length; i++) {
			minDist[i] = StdStats.min(distMatrix[i]);
		}

		double mean = StdStats.mean(minDist);
		double std = StdStats.stddev(minDist);
		double jumpcosts = mean + std;
		return jumpcosts <= MIN_JUMPCOSTS ? MIN_JUMPCOSTS : jumpcosts;
	}

	private static double[][] extendMatrix(double[][] matx, int m, int n) {
		double[][] matxE = new double[m + 2][n + 2];
		for (int i = 0; i < m + 2; i++) {
			Arrays.fill(matxE[i], Double.MAX_VALUE);
		}
		for (int i = 0; i < m; i++) {
			for (int j = 0; j < n; j++) {
				matxE[i + 1][j + 1] = matx[i][j];
			}
		}
		matxE[0][0] = 0;
		matxE[m + 1][n + 1] = 0;

		return matxE;
	}

	public static void main(String[] args) throws Exception {
		double[] t1 = new double[] { 0.126, 0.376, 0.626, 0.751, 0.876 };
		double[] t2 = new double[] { 0.026, 0.076, 0.226, 0.476, 0.726, 0.851, 0.976 };

		double[] s1 = new double[] { 0.01234234, 0.01209678, 0.02221759, 0.01248292, 0.01249573 };
		double[] s2 = new double[] { 0.01124079, 0.01031078, 0.01234234, 0.01209678, 0.02221759, 0.01248292, 0.01249572 };

		System.out.println(osb(t1, t2, s1, s2, 3, 5, 5, 0.1107));
	}
}
